@@ -26,10 +26,14 @@ import warnings
 
 from math import log10
 from pkg_resources import resource_filename
+from copy import deepcopy
 
-from .utils import (STR2BYTES, BYTES2STR, NULLCHAR_2,
-                    strjoin, memcopy, is_string, is_string_or_bytes,
-                    ascii_string, clib_search_path)
+from .utils import (str2bytes, bytes2str, strjoin, EPICS_STR_ENCODING,
+                    clib_search_path)
+
+# legacy imports in case someone was importing them from here..
+from .utils import (STR2BYTES, BYTES2STR, NULLCHAR_2, memcopy, is_string,
+                    is_string_or_bytes, ascii_string)
 
 # ignore warning about item size... for now??
 warnings.filterwarnings('ignore',
@@ -65,6 +69,9 @@ PREEMPTIVE_CALLBACK = True
 
 AUTO_CLEANUP = True
 
+# A sentinel to mark libca as going through the shutdown process
+_LIBCA_FINALIZED = object()
+
 ##
 # maximum element count for auto-monitoring of PVs in epics.pv
 # and for automatic conversion of numerical array data to numpy arrays
@@ -348,14 +355,14 @@ def initialize_libca():
     libca : object
         ca library object, used for all subsequent ca calls
 
-    See Also
-    --------
-    withCA :  decorator to ensure CA is initialized
+
 
     Notes
     -----
     This function must be called prior to any real CA calls.
 
+    See the `withCA`  decorator to ensure CA is initialized
+
     """
     if 'EPICS_CA_MAX_ARRAY_BYTES' not in os.environ:
         os.environ['EPICS_CA_MAX_ARRAY_BYTES'] = "%i" %  2**24
@@ -398,13 +405,7 @@ def initialize_libca():
     # save value offests used for unpacking
     # TIME and CTRL data as an array in dbr module
 
-    # in_dll is not available for arrays in IronPython, so use a reference to the first element
-    if dbr.IRON_PYTHON:
-        value_offset0 = ctypes.c_short.in_dll(libca,'dbr_value_offset')
-        dbr.value_offset = ctypes.cast(ctypes.addressof(value_offset0),
-                                       (39*ctypes.c_short))
-    else:
-        dbr.value_offset = (39*ctypes.c_short).in_dll(libca,'dbr_value_offset')
+    dbr.value_offset = (39*ctypes.c_short).in_dll(libca,'dbr_value_offset')
 
     initial_context = current_context()
     if AUTO_CLEANUP:
@@ -424,7 +425,7 @@ def finalize_libca(maxtime=10.0):
 
     """
     global libca
-    if libca is None:
+    if libca is None or libca is _LIBCA_FINALIZED:
         return
     try:
         start_time = time.time()
@@ -446,7 +447,7 @@ def finalize_libca(maxtime=10.0):
             poll()
             flush_count += 1
         context_destroy()
-        libca = None
+        libca = _LIBCA_FINALIZED
     except:
         pass
     time.sleep(0.01)
@@ -466,7 +467,7 @@ def _get_cache_by_chid(chid):
         # context cache before giving up. This branch should not happen often.
         context = current_context()
         if context is not None:
-            pvname = BYTES2STR(libca.ca_name(dbr.chid_t(chid)))
+            pvname = bytes2str(libca.ca_name(dbr.chid_t(chid)))
             return _cache[context][pvname]
         raise
 
@@ -500,6 +501,12 @@ def clear_cache():
     multiprocessing (and is done internally by CAProcess),
     but can be useful to fully reset a Channel Access session.
     """
+    # Unregister callbacks (if any)
+    for chid, entry in list(_chid_cache.items()):
+        try:
+            clear_channel(chid)
+        except ChannelAccessException:
+            pass
 
     # Clear global state variables
     _cache.clear()
@@ -541,6 +548,8 @@ def withCA(fcn):
         global libca
         if libca is None:
             initialize_libca()
+        elif libca is _LIBCA_FINALIZED:
+            return  # Avoid raising exceptions when Python shutting down
         return fcn(*args, **kwds)
     return wrapper
 
@@ -556,6 +565,10 @@ def withCHID(fcn):
     @functools.wraps(fcn)
     def wrapper(*args, **kwds):
         "withCHID wrapper"
+        global libca
+        if libca is _LIBCA_FINALIZED:
+            return  # Avoid raising exceptions when Python shutting down
+
         if len(args)>0:
             chid = args[0]
             args = list(args)
@@ -580,6 +593,10 @@ def withConnectedCHID(fcn):
     @functools.wraps(fcn)
     def wrapper(*args, **kwds):
         "withConnectedCHID wrapper"
+        global libca
+        if libca is _LIBCA_FINALIZED:
+            return  # Avoid raising exceptions when Python shutting down
+
         if len(args)>0:
             chid = args[0]
             args = list(args)
@@ -606,6 +623,10 @@ def withMaybeConnectedCHID(fcn):
     @functools.wraps(fcn)
     def wrapper(*args, **kwds):
         "withMaybeConnectedCHID wrapper"
+        global libca
+        if libca is _LIBCA_FINALIZED:
+            return  # Avoid raising exceptions when Python shutting down
+
         if len(args)>0:
             chid = args[0]
             args = list(args)
@@ -676,6 +697,10 @@ def _onMonitorEvent(args):
         return
 
     value = dbr.cast_args(args)
+    if value[1] is None:
+        # Cannot process the input becaue casting failed.
+        return
+
     kwds = {'ftype':args.type, 'count':args.count,
             'chid': args.chid, 'pvname': entry.pvname}
 
@@ -715,19 +740,21 @@ def _onGetEvent(args, **kws):
     except KeyError:
         return
 
-    ftype = (args.usr.value if dbr.IRON_PYTHON
-             else args.usr)
-
+    ftype = args.usr
     if args.status != dbr.ECA_NORMAL:
         result = ChannelAccessGetFailure(
             'Get failed; status code: %d' % args.status,
             chid=args.chid,
             status=args.status
         )
-    elif dbr.IRON_PYTHON:
-        result = dbr.cast_args(args)
     else:
-        result = memcopy(dbr.cast_args(args))
+        result = deepcopy(dbr.cast_args(args))
+        if result[1] is None:
+            result = ChannelAccessGetFailure(
+                'Get failed; unknown type: %d' % args.type,
+                chid=args.chid,
+                status=args.status
+            )
 
     with entry.lock:
         entry.get_results[ftype][0] = result
@@ -816,14 +843,12 @@ def attach_context(context):
 @withCA
 @withSEVCHK
 def use_initial_context():
-    """Attaches to the context created when libca is initialized.
+    """
+    Attaches to the context created when libca is initialized.
     Using this function is recommended when writing threaded programs that
     using CA.
 
-    See Also
-    --------
-    :ref:`advanced-threads-label` in doc for further discussion.
-
+    See the advanced section in doc for further discussion.
     """
     global initial_context
     ret = dbr.ECA_NORMAL
@@ -868,13 +893,13 @@ def flush_io():
 def message(status):
     """Print a message corresponding to a Channel Access status return value.
     """
-    return BYTES2STR(libca.ca_message(status))
+    return bytes2str(libca.ca_message(status))
 
 @withCA
 def version():
     """   Print Channel Access version string.
     Currently, this should report '4.13' """
-    return BYTES2STR(libca.ca_version())
+    return bytes2str(libca.ca_version())
 
 @withCA
 def pend_io(timeout=1.0):
@@ -980,7 +1005,7 @@ def create_channel(pvname, connect=False, auto_cb=True, callback=None):
             chid = dbr.chid_t()
             with entry.lock:
                 ret = libca.ca_create_channel(
-                    ctypes.c_char_p(STR2BYTES(pvname)), _CB_CONNECT, 0, 0,
+                    ctypes.c_char_p(str2bytes(pvname)), _CB_CONNECT, 0, 0,
                     ctypes.byref(chid)
                 )
                 PySEVCHK('create_channel', ret)
@@ -1090,12 +1115,12 @@ def _chid_to_int(chid):
 @withCHID
 def name(chid):
     "return PV name for channel name"
-    return BYTES2STR(libca.ca_name(chid))
+    return bytes2str(libca.ca_name(chid))
 
 @withCHID
 def host_name(chid):
     "return host name and port serving Channel"
-    return BYTES2STR(libca.ca_host_name(chid))
+    return bytes2str(libca.ca_host_name(chid))
 
 @withCHID
 def element_count(chid):
@@ -1197,9 +1222,9 @@ def _unpack(chid, data, count=None, ftype=None, as_numpy=True):
         """ Scan a string, or an array of strings as a list, depending on content """
         out = []
         for elem in range(min(count, len(data))):
-            this = strjoin('', BYTES2STR(data[elem].value)).rstrip()
-            if NULLCHAR_2 in this:
-                this = this[:this.index(NULLCHAR_2)]
+            this = strjoin('', bytes2str(data[elem].value)).rstrip()
+            if '\x00' in this:
+                this = this[:this.index('\x00')]
             out.append(this)
         if len(out) == 1:
             out = out[0]
@@ -1213,9 +1238,9 @@ def _unpack(chid, data, count=None, ftype=None, as_numpy=True):
                 out = numpy.empty(shape=(count,), dtype=dbr.NP_Map[ntype])
                 ctypes.memmove(out.ctypes.data, data, out.nbytes)
             else:
-                out = numpy.ctypeslib.as_array(memcopy(data))
+                out = numpy.ctypeslib.as_array(deepcopy(data))
         else:
-            out = memcopy(data)
+            out = deepcopy(data)
         return out
 
     def unpack(data, count, ntype, use_numpy, elem_count):
@@ -1285,10 +1310,10 @@ def _unpack_metadata(ftype, dbr_value):
             if hasattr(dbr_value, attr):
                 md[attr] = getattr(dbr_value, attr)
                 if attr == 'units':
-                    md[attr] = BYTES2STR(getattr(dbr_value, attr, None))
+                    md[attr] = bytes2str(getattr(dbr_value, attr, None))
 
         if hasattr(dbr_value, 'strs') and getattr(dbr_value, 'no_str', 0) > 0:
-            md['enum_strs'] = tuple(BYTES2STR(dbr_value.strs[i].value)
+            md['enum_strs'] = tuple(bytes2str(dbr_value.strs[i].value)
                                     for i in range(dbr_value.no_str))
     elif ftype >= dbr.TIME_STRING:
         md['status'] = dbr_value.status
@@ -1338,9 +1363,7 @@ def get_with_metadata(chid, ftype=None, count=None, wait=True, timeout=None,
        Returns ``None`` if the channel is not connected, `wait=False` was used,
        or the data transfer timed out.
 
-    See also
-    --------
-    See :func:`get` for additional usage notes.
+    See `get()` for additional usage notes.
     """
     if ftype is None:
         ftype = field_type(chid)
@@ -1474,9 +1497,7 @@ def get_complete_with_metadata(chid, ftype=None, count=None, timeout=None,
        This function will return ``None`` if the previous :func:`get` actually
        completed, or if this data transfer also times out.
 
-    See also
-    --------
-    See :func:`get_complete` for additional usage notes.
+    See `get_complete()` for additional usage notes.
     """
     if ftype is None:
         ftype = field_type(chid)
@@ -1526,7 +1547,7 @@ def get_complete_with_metadata(chid, ftype=None, count=None, timeout=None,
     if as_string:
         val = _as_string(val, chid, count, ftype)
     elif isinstance(val, ctypes.Array) and HAS_NUMPY and as_numpy:
-        val = numpy.ctypeslib.as_array(memcopy(val))
+        val = numpy.ctypeslib.as_array(deepcopy(val))
 
     # value retrieved, clear cached value
     metadata['value'] = val
@@ -1595,7 +1616,7 @@ def _as_string(val, chid, count, ftype):
 
 @withConnectedCHID
 def put(chid, value, wait=False, timeout=30, callback=None,
-        callback_data=None):
+        callback_data=None, ftype=None):
     """sets the Channel to a value, with options to either wait
     (block) for the processing to complete, or to execute a
     supplied callback function when the process has completed.
@@ -1614,6 +1635,8 @@ def put(chid, value, wait=False, timeout=30, callback=None,
         user-supplied function to run when processing has completed.
     callback_data :  object
         extra data to pass on to a user-supplied callback function.
+    ftype : ``None`` or int (valid dbr type)
+        force field type to be a non-native form (None will use native form)
 
     Returns
     -------
@@ -1628,12 +1651,13 @@ def put(chid, value, wait=False, timeout=30, callback=None,
         pvname=pvname, data=callback_data
 
     """
-    ftype = field_type(chid)
+    if ftype is None:
+        ftype = field_type(chid)
     count = nativecount = element_count(chid)
     if count > 1:
         # check that data for array PVS is a list, array, or string
         try:
-            if ftype == dbr.STRING and is_string_or_bytes(value):
+            if ftype == dbr.STRING and isinstance(value, (str, bytes)):
                 # len('abc') --> 3, however this is one element for dbr.STRING ftype
                 count = 1
             else:
@@ -1644,24 +1668,24 @@ def put(chid, value, wait=False, timeout=30, callback=None,
         except TypeError:
             write('''PyEpics Warning:
      value put() to array PV must be an array or sequence''')
-    if ftype == dbr.CHAR and nativecount > 1 and is_string_or_bytes(value):
+    if ftype == dbr.CHAR and nativecount > 1 and isinstance(value, (str, bytes)):
         count += 1
         count = min(count, nativecount)
 
-    # if needed (python3, especially) convert to basic string/bytes form
-    if is_string(value):
-        value = ascii_string(value)
+    # if needed convert to basic string/bytes git stform
+    if isinstance(value, str):
+        value = bytes(value, EPICS_STR_ENCODING)
 
     data = (count*dbr.Map[ftype])()
     if ftype == dbr.STRING:
-        if is_string_or_bytes(value):
+        if isinstance(value, (str, bytes)):
             data[0].value = value
         else:
             for elem in range(min(count, len(value))):
-                data[elem].value = ascii_string(value[elem])
+                data[elem].value = bytes(str(value[elem]), EPICS_STR_ENCODING)
     elif nativecount == 1:
         if ftype == dbr.CHAR:
-            if is_string_or_bytes(value):
+            if isinstance(value, (str, bytes)):
                 if isinstance(value, bytes):
                     value = value.decode('ascii', 'replace')
                 value = [ord(i) for i in value] + [0, ]
@@ -1669,7 +1693,7 @@ def put(chid, value, wait=False, timeout=30, callback=None,
                 data[0] = value
         else:
             # allow strings (even bits/hex) to be put to integer types
-            if is_string(value) and isinstance(data[0], (int, )):
+            if isinstance(value, (str, bytes)) and isinstance(data[0], (int, )):
                 value = int(value, base=0)
             try:
                 data[0] = value
@@ -1681,7 +1705,7 @@ def put(chid, value, wait=False, timeout=30, callback=None,
                 raise ChannelAccessException(errmsg % (repr(value), tname))
 
     else:
-        if ftype == dbr.CHAR and is_string_or_bytes(value):
+        if ftype == dbr.CHAR and isinstance(value, (str, bytes)):
             if isinstance(value, bytes):
                 value = value.decode('ascii', 'replace')
             value = [ord(i) for i in value] + [0, ]
@@ -1983,9 +2007,9 @@ def sg_put(gid, chid, value):
     else:
         # auto-convert strings to arrays for character waveforms
         # could consider using
-        # numpy.fromstring(("%s%s" % (s,NULLCHAR*maxlen))[:maxlen],
+        # numpy.fromstring(("%s%s" % (s, pythonb'\x00'*maxlen))[:maxlen],
         #                  dtype=numpy.uint8)
-        if ftype == dbr.CHAR and is_string_or_bytes(value):
+        if ftype == dbr.CHAR and isinstance(value, (str, bytes)):
             pad = [0]*(1+count-len(value))
             if isinstance(value, bytes):
                 value = value.decode('ascii', 'replace')
