@@ -15,7 +15,6 @@ from math import log10
 
 from . import ca
 from . import dbr
-from .utils import is_string
 
 try:
     from types import SimpleNamespace as Namespace
@@ -319,7 +318,6 @@ class PV(object):
             self._args['type'] = _ftype_
             self._args['typefull'] = _ftype_
             self._args['ftype'] = dbr.Name(_ftype_, reverse=True)
-            self._check_auto_monitor()
 
         for conn_cb in self.connection_callbacks:
             if callable(conn_cb):
@@ -334,6 +332,8 @@ class PV(object):
         # threads from thinking a connection is complete when it is actually
         # still in progress.
         self.connected = conn
+        if conn:
+            self._check_auto_monitor()
 
     @_ensure_context
     def _clear_auto_monitor_subscription(self):
@@ -354,10 +354,13 @@ class PV(object):
 
         Clears or adds monitor, if necessary.
         '''
-        count = self.count
-        chid = self.chid
+        if not self.connected or self.chid is None:
+            # Auto-monitor will be enabled (or toggled based on count) upon the
+            # next connection callback.
+            return
 
-        if count is None or chid is None:
+        count = self.count
+        if count is None:
             return
 
         if self._auto_monitor is None:
@@ -626,7 +629,7 @@ class PV(object):
             return None
 
         if (self.ftype in (dbr.ENUM, dbr.TIME_ENUM, dbr.CTRL_ENUM) and
-            is_string(value)):
+            isinstance(value, str)):
             if self._args['enum_strs'] is None:
                 self.get_ctrlvars()
             if value in self._args['enum_strs']:
@@ -827,9 +830,13 @@ class PV(object):
             self.callbacks.pop(index)
             ca.poll()
 
-    def clear_callbacks(self):
+    def clear_callbacks(self, with_access_callback=False, with_connect_callback=False):
         "clear all callbacks"
         self.callbacks.clear()
+        if with_access_callback:
+            self.access_callbacks = []
+        if with_connect_callback:
+            self.connection_callbacks = []
 
     def _getinfo(self):
         "get information paragraph"
@@ -863,8 +870,8 @@ class PV(object):
             out.append("   value      = array  [%s%s]" % (",".join(aval), ext))
         for nam in ('char_value', 'count', 'nelm', 'type', 'units',
                     'precision', 'host', 'access',
-                    'status', 'severity', 'timestamp',
-                    'posixseconds', 'nanoseconds',
+                    'status', 'char_status', 'severity', 'char_severity',
+                    'timestamp', 'posixseconds', 'nanoseconds',
                     'upper_ctrl_limit', 'lower_ctrl_limit',
                     'upper_disp_limit', 'lower_disp_limit',
                     'upper_alarm_limit', 'lower_alarm_limit',
@@ -930,13 +937,19 @@ class PV(object):
     @property
     def char_value(self):
         "character string representation of value"
-        return self._getarg('char_value')
+        self._getarg('char_value')  # forces lookup of CTRL vars
+        return self._set_charval(self._getarg('value'))
 
     @property
     def status(self):
         "pv status"
         return self._getarg('status')
 
+    @property
+    def char_status(self):
+        "character string representation of the pv status"
+        return dbr.AlarmStatus(self.status).name
+
     @property
     def type(self):
         "pv type"
@@ -991,6 +1004,11 @@ class PV(object):
         "pv severity"
         return self._getarg('severity')
 
+    @property
+    def char_severity(self):
+        "character string representation of the pv severity"
+        return dbr.AlarmSeverity(self.severity).name
+
     @property
     def timestamp(self):
         "timestamp of last pv action"
@@ -1091,8 +1109,22 @@ class PV(object):
             return False
 
     @_ensure_context
-    def disconnect(self):
-        "disconnect PV"
+    def disconnect(self, deepclean=True):
+        """
+        disconnect PV
+
+        this method clears all the user-defined callbacks for a PV and removes
+        it from _PVcache_, so that subsequent connection to this PV will almost
+        always make a completely new connection.
+
+        Arguments
+        -----------
+        deepclean, bool  removal all cache connection and access-rights callbacks [True]
+
+
+        With deepclean=False, references to callbacks for connection and access-rights
+        events will not be removed from the ca _cache for the current context.
+        """
         self.connected = False
 
         ctx = ca.current_context()
@@ -1100,18 +1132,25 @@ class PV(object):
         if pvid in _PVcache_:
             _PVcache_.pop(pvid)
 
-        cache_item = ca._cache[ctx].pop(self.pvname, None)
-        if cache_item is not None:
-            if self._monref is not None:
-                # atexit may have already cleared the subscription
-                self._clear_auto_monitor_subscription()
+        if deepclean:
+            cache_item = ca._cache[ctx].get(self.pvname, None)
+            if cache_item is not None:
+                # removing all references from ca._cache to this object,
+                # so that it can be garbage-collected
+                for cb_cache, cb_ref in ((cache_item.callbacks,
+                                          self.__on_connect),
+                                         (cache_item.access_event_callback,
+                                          self.__on_access_rights_event)):
+                    while cb_cache.count(cb_ref) > 0:
+                        cb_cache.remove(cb_ref)
 
-            # TODO: clear channel should be called as well
-            # ca.clear_channel(cache_item.chid)
+        if self._monref is not None:
+            # atexit may have already cleared the subscription
+            self._clear_auto_monitor_subscription()
 
         self._monref = None
         self._monref_mask = None
-        self.clear_callbacks()
+        self.clear_callbacks(True, True)
         self._args = {}.fromkeys(self._fields)
         ca.poll(evt=1.e-3, iot=1.0)
 
